from os import path
import numpy as np

PATH = path.abspath(path.dirname(__file__))

def smallGaussian():
    """
        Relaxed 10 body, gaussian distribution initial conditions.

        `G = 1.0`

        `mass_unit = 1.0`

        Returns:
            np.ndarray: contains points properties.
    """
    return np.genfromtxt(path.join(PATH, "small_size_gaussian2d.dat"))

def mediumGaussian():
    """
        Relaxed 100 body, gaussian distribution initial conditions.

        `G = 1.0`

        `mass_unit = 1.0`

        Returns:
            np.ndarray: contains points properties.
    """
    return np.genfromtxt(path.join(PATH, "medium_size_gaussian2d.dat"))

def largeGaussian():
    return np.genfromtxt(path.join(PATH, "large_size_gaussian2d.dat"))

def twosmallGaussians():
    return np.genfromtxt(path.join(PATH, "two_small_size_gaussian2d.dat"))

def twomediumGaussians():
    return np.genfromtxt(path.join(PATH, "two_medium_size_gaussian2d.dat"))

def twolargeGaussians():
    return np.genfromtxt(path.join(PATH, "two_large_size_gaussian2d.dat"))



import os
import sys
from ctypes import c_double, c_float, CDLL

DOUBLE = c_float #: floating point precision

NAME = "astrohutc.cpython-%d%dm.so"%sys.version_info[:2]
PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), NAME) #: path to the C shared library

LIB = CDLL(PATH) #: CDLL instance of the shared library



import os
import sys
import ctypes
import numpy as np
from .constants import DOUBLE, LIB
from .structs2d import body2d, node2d, point2d

def fromBodiesToArray(bodies, N, dim = 2):
    array = np.zeros((N, dim*3 + 1))
    for i in range(N):
        array[i] = bodies[i].asList()

    return array

def fromArrayToBodies(array, dim = 2):
    rows, cols = array.shape[:2]
    bodies = LIB.malloc(rows * ctypes.sizeof(body2d))
    bodies = ctypes.cast(bodies, ctypes.POINTER(body2d))

    if dim == 2:
        body = body2d
    elif dim == 3:
        body = body3d

    if cols//dim == 3:
        for i in range(rows):
            bodies[i] = body2d(point2d(*array[i, :2]), point2d(*array[i, 2:4]),
                                point2d(*array[i, 4:]))
    elif cols//dim == 2:
        for i in range(rows):
            bodies[i] = body2d(point2d(*array[i, :2]), point2d(*array[i, 2:4]))
    else:
        raise(Exception("Input file has wrong data format."))
    return rows, bodies

def fromNodeToArray(node, dim = 2):
    def internalNode(node, dim):
        try:
            if node.contents.Nbodies == 1:
                c1 = point2d()
                c2 = point2d()
                c3 = point2d()
                c4 = point2d()

                c1.x = node.contents.center.x - node.contents.width*0.5
                c1.y = node.contents.center.y + node.contents.height*0.5

                c2.x = node.contents.center.x + node.contents.width*0.5
                c2.y = node.contents.center.y + node.contents.height*0.5

                c3.x = node.contents.center.x + node.contents.width*0.5
                c3.y = node.contents.center.y - node.contents.height*0.5

                c4.x = node.contents.center.x - node.contents.width*0.5
                c4.y = node.contents.center.y - node.contents.height*0.5

                return [node.contents.xs[0], node.contents.ys[0], c1.x, c2.x, c3.x, c4.x, c1.x,
                        c1.y, c2.y, c3.y, c4.y, c1.y]
        except ValueError:
            return None

        else:
            answer = []
            if dim == 2:
                nodes = [node.contents.subnode1, node.contents.subnode2, node.contents.subnode3, node.contents.subnode4]
                for item in nodes:
                    ans = internalNode(item, dim)
                    if ans != None:
                        answer += ans
            else:
                raise(Exception("Not implemented yet"))

            return answer

    answer = internalNode(node, dim)
    answer = np.array(answer).reshape(node.contents.Nbodies, len(answer)//node.contents.Nbodies)

    return answer

def generateSpeeds(positions, G, mass_unit):
    """
        Generates rotation speeds based on positions, G constants and mass_unit value.

        Returns:
            np.ndarray: array like positions.
    """
    N, dim = positions.shape

    xs = positions[:, 0]
    ys = positions[:, 1]

    x0 = xs.mean()
    y0 = ys.mean()

    xs = xs - x0
    ys = ys - y0

    if dim == 3:
        zs = positions[:, 2]
        z0 = zs.mean()

    speeds = np.zeros_like(positions)

    if dim == 2:
        d = np.sqrt(xs**2 + ys**2)
        for i in range(N):
            n = sum(d < d[i])
            mag = 0.9*np.sqrt(G*n*mass_unit/d[i])
            speeds[i, 0] = -ys[i]*mag/d[i]
            speeds[i, 1] = xs[i]*mag/d[i]

    return speeds

def createArray(pos, speeds, acc = None):
    """
        Creates and array containing the whole properties of the system.

        np.ndarray: appends pos, speeds and acc if any.
    """
    N, dim = pos.shape
    array = np.zeros((N, dim*3))

    array[:, :dim] = pos
    array[:, dim:2*dim] = speeds

    if type(acc) != type(None):
        array[:, 2*dim:] = acc

    return array



import ctypes
from .constants import DOUBLE

class point2d(ctypes.Structure):
    """
        Defines a point2d structure.
    """
    _fields_ = [('x', DOUBLE),
                ('y', DOUBLE)]

    def __str__(self):
        return "x = %f, y = %f"%(self.x, self.y)

    def getX(self):
        return self.x

class body2d(ctypes.Structure):
    """
        Defines a body2d structure.
    """
    _fields_ = [('p', point2d),
                ('v', point2d),
                ('a', point2d),
                ('E', DOUBLE)]

    def __str__(self):
        values = ["%s: %s"%(item, str(getattr(self, item))) for item, _ in self._fields_]
        return "\n".join(values)

    def asList(self):
        return [self.p.x, self.p.y, self.v.x, self.v.y, self.a.x, self.a.y, self.E]

    def asarray(self):
        return np.array(self.asList())

class node2d(ctypes.Structure):
    """
        Defines a node2d structure.
    """
    def __str__(self):
        toprint = ["Nbodies", "mass", "width", "height", "center", "cmass"]
        values = ["%s: %s"%(item, str(getattr(self, item))) for item in toprint]
        return "\n".join(values)

node2d._fields_ = [('xs', ctypes.POINTER(DOUBLE)),
        ('ys', ctypes.POINTER(DOUBLE)),

        ('subnode1', ctypes.POINTER(node2d)),
        ('subnode2', ctypes.POINTER(node2d)),
        ('subnode3', ctypes.POINTER(node2d)),
        ('subnode4', ctypes.POINTER(node2d)),

        ('Nbodies', ctypes.c_int),
        ('mass', DOUBLE),
        ('width', DOUBLE),
        ('height', DOUBLE),
        ('cmass', point2d),
        ('center', point2d)]



import numpy as np
import astrohut as ah
import matplotlib.pyplot as plt

G = 1
m = 1

pos = np.random.normal(size=(100, 2))
speeds = ah.generateSpeeds(pos, G, m)

system = ah.createArray(pos, speeds)

sim = ah.Simulation(system, tau = 1.0, epsilon = 1e-5)

# nInstants = int(sim.calcRelaxationTime()/sim.dt)
nInstants = 10000

sim.start(nInstants, save_to_array_every = nInstants//100, save_to_file_every = nInstants)
ani = sim.makeAnimation()

plt.show()

plt.plot(sim.getEnergies())
plt.show()



import astrohut as ah
from setuptools import setup
from setuptools.extension import Extension

module = Extension('astrohut/core/astrohutc',
                    sources = ['astrohut/core/constants.c', 'astrohut/core/common.c', 'astrohut/core/box2d.c', 'astrohut/core/box3d.c'], include_dirs=['astrohut/core'], extra_compile_args=["-fopenmp", "-O2"],
                     extra_link_args=["-fopenmp", "-O2"])

setup(
    name = "astrohut",
    version = ah.__version__,
    author = "Juan Barbosa",
    author_email = "js.barbosa10@uniandes.edu.co",
    description = ('Barnes-Hut NBody simulation library.'),
    license = "GPL",
    keywords = "example documentation tutorial",
    packages=['astrohut', 'astrohut/core', 'astrohut/data', 'astrohut/testing'],
    install_requires=['matplotlib', 'numpy'],
    ext_modules = [module],
    long_description="https://github.com/jsbarbosa/astrohut/",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License (GPL)",
    ],
    headers = ['astrohut/core/box3d.h', 'astrohut/core/box2d.h', 'astrohut/core/common.h', 'astrohut/core/constants.h'],
    include_package_data = True,
)



import os
import ctypes
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from ctypes import POINTER, c_char_p, c_int, pointer

from .constants import DOUBLE, LIB
from .structs2d import body2d, node2d
from .core import fromArrayToBodies, fromBodiesToArray, fromNodeToArray

LIB.malloc.restype = ctypes.c_void_p
LIB.calloc.restype = ctypes.c_void_p

LIB.setConstants.argtypes = DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE

LIB.loadFile2d.argtypes = c_char_p, c_char_p, c_int
LIB.loadFile2d.restype = POINTER(body2d)

LIB.initFirstNode2d.argtypes = c_int, POINTER(body2d)
LIB.initFirstNode2d.restype = POINTER(node2d)

LIB.solveInterval2d.argtypes = c_int, POINTER(POINTER(node2d)), POINTER(body2d)
LIB.solveInterval2d.restype = POINTER(body2d)

LIB.solveInstant2d.argtypes = POINTER(POINTER(node2d)), POINTER(body2d)
LIB.solveInstant2d.restype = POINTER(body2d)

LIB.setPrint.argtypes = c_char_p, c_int

LIB.omp_set_num_threads.argtypes = c_int,

def __setPrint__(prefix = "", frames_every = 0):
    LIB.setPrint(prefix.encode(), frames_every)

def __setConstants__(mass_unit, G, tau, dt, epsilon):
    LIB.setConstants(mass_unit, G, tau, dt, epsilon)

class Simulation():
    """
    Main class of astrohut. In order to evolve a system and study them a simulation instance
    must be launch.
    """
    def __init__(self, data, dim = 2, mass_unit = 1.0, G = 1.0, tau = 0.5, dt = 1e-4, epsilon = 1e-4,
                read_kwargs = {}):
        self.data = data
        self.dim = dim
        self.mass_unit = mass_unit
        self.G = G
        self.tau = tau
        self.dt = dt
        self.epsilon = epsilon

        if type(self.data) is str:
            self.data = np.genfromtxt(data, **read_kwargs)

        self.Nbodies, self.bodies = fromArrayToBodies(self.data, self.dim)
        self.node = None
        self.results_nodes = None
        self.results_bodies = None
        self.setConstants()
        self.setFilePrefix()

        self.ax = None
        self.fig = None

        self.file_number = 0

    def setConstants(self):
        """
            Sets the current constants at the C shared library.
        """
        __setConstants__(self.mass_unit, self.G, self.tau, self.dt, self.epsilon)

    def setFilePrefix(self, prefix = ""):
        """
            Sets the file prefix at the C shared library.
        """
        __setPrint__(os.path.join(os.getcwd(), prefix))

    def start(self, Ninstants, threads = 0, save_to_file_every = 0, save_to_array_every = 1):
        """
            Starts a simulation, evolves the system N Ninstants of time.

            Returns:
                np.ndarray: 3d array of the form (instants, Nbodies, properties).
                        Properties are positions, speeds, accelerations and energy.
                        Non scalar properties are arranged as components.
                np.ndarray: 3d array of the form (instants, Nboxes, (center, box)).
                        For every instant there are Nboxes rows, describing each box.
                        First two or three columns are the coordinates of the center
                        of the box. Next columns describe the corners of the box,
                        first the x components, then y, and finally z if dim is 3.

        """
        if type(self.node) == type(None):
            self.node = LIB.initFirstNode2d(self.Nbodies, self.bodies)

        if threads > 0:
            LIB.omp_set_num_threads(threads)

        Ninstants = int(Ninstants)
        new = LIB.solveInstant2d(ctypes.byref(self.node), self.bodies)
        array_number = 0

        if save_to_array_every != 0:
            instant_points = np.zeros((Ninstants//save_to_array_every + 1, self.Nbodies, 3*self.dim + 1))
            instant_nodes = np.zeros((Ninstants//save_to_array_every + 1, self.Nbodies, 12))

        for i in range(Ninstants + 1):
            new2 = LIB.solveInstant2d(ctypes.byref(self.node), new)

            if save_to_file_every > 0:
                if i%save_to_file_every == 0:
                    LIB.printInstant2d(self.node, new, self.file_number)
                    self.file_number += 1

            if save_to_array_every > 0:
                if i%save_to_array_every == 0:
                    instant_points[array_number] = fromBodiesToArray(new, self.Nbodies)
                    instant_nodes[array_number] = fromNodeToArray(self.node, self.dim)
                    array_number += 1

            LIB.swapBody2d(ctypes.byref(new2), ctypes.byref(new))
            LIB.free(new2)

        if save_to_array_every != 0:
            if type(self.results_bodies) == type(None):
                self.results_bodies = instant_points
            if type(self.results_nodes) == type(None):
                self.results_nodes = instant_nodes
            else:
                self.results_bodies = np.vstack((self.results_bodies, instant_points))
                self.results_nodes = np.vstack((self.results_nodes, instant_nodes))
            return self.results_bodies, self.results_nodes
        else:
            return fromBodiesToArray(new, self.Nbodies), fromNodeToArray(self.node)

    def animate2d(self, i, values, points, boxes):
        """
            Method to be used with matplotlib FuncAnimation.
        """
        nboxes = len(boxes)
        npoints = len(points)
        if len(boxes) > 0:
            for j in range(nboxes):
                box = values[i, j, 2:]
                boxes[j].set_data(box[:5], box[5:])

        for j in range(npoints):
            body = values[i, j]
            points[j].set_data(body[0], body[1])

        if nboxes > 0:
            return points, boxes
        else:
            return points,

    def configPlot(self, figsize=(6, 4.5), xlabel = None, ylabel = None):
        """
            Configures the plot.

            Returns:
                matplotlib.figure: figure containing the main plot.
                matplotlib.axes: axes containing all the dots and boxes frames (if boxed).
        """
        self.fig, self.ax = plt.subplots(figsize = figsize)

        if xlabel == None:
            self.ax.set_xlabel("$x$")

        if ylabel == None:
            self.ax.set_xlabel("$y$")

        return self.fig, self.ax

    def makeAnimation(self, boxed = False, color = None, alpha = 0.5):
        """
            Makes an animation of `data`.

            Args:
                boxed (bool): If True draws the boxes frames.

                color (string):If None, multiplecolors are used. If string is passed a monochrome color is used.

                alpha (float): Alpha channel argument.

            Returns:
                matplotlib.animation.FuncAnimation: animation of the data.
        """
        if boxed:
            data_to_use = self.results_nodes
        else:
            data_to_use = self.results_bodies

        if type(data_to_use) == type(None):
            self.start(100)

        data = data_to_use

        Ninstants = data.shape[0]

        if self.ax == None or self.fig == None:
            self.configPlot()

        if color == None:
            points = [self.ax.plot([], [], "o", alpha = alpha)[0] for i in range(self.Nbodies)]
        else:
            points = [self.ax.plot([], [], "o", color = color, alpha = alpha)[0] for i in range(self.Nbodies)]
        boxes = []

        if boxed:
            boxes = [self.ax.plot([], [], c = points[i].get_color(), alpha = alpha)[0] for i in range(self.Nbodies)]

        self.ax.set_xlim(data[:, :, 0].min(), data[:, :, 0].max())
        self.ax.set_ylim(data[:, :, 1].min(), data[:, :, 1].max())

        if self.dim == 2:
            animation = self.animate2d

        ani = FuncAnimation(self.fig, animation, frames = Ninstants,
                            interval = 25, fargs=(data, points, boxes))

        return ani

    def getEnergies(self):
        """
            Returns:
                np.ndarray: contains the energy of the system at every saved instant of time.
        """
        if type(self.results_bodies) != type(None):
            return self.results_bodies[:, :, -1].sum(axis=1)
        else:
            raise(Exception("No simulation has been started."))

    def calcRelaxationTime(self):
        """
            Calculates the relaxation time required for the simulation to reach equilibrium.

            Returns:
                float: relaxation time of the system.
        """
        if self.dim == 2:
            r = ((self.data[:, 0] - self.data[:, 0].mean())**2 + \
                (self.data[:, 1] - self.data[:, 1].mean())**2)**0.5

        elif self.dim == 3:
            r = ((self.data[:, 0] - self.data[:, 0].mean())**2 + \
                (self.data[:, 1] - self.data[:, 1].mean())**2 + \
                (self.data[:, 2] - self.data[:, 2].mean())**2)**0.5

        n = self.Nbodies/(r.max() - r.min())**3
        tcr = 1/np.sqrt(self.G*self.mass_unit*n)

        return self.Nbodies/(10*np.log10(self.Nbodies)) * tcr



import numpy as np
import astrohut as ah
import matplotlib.pyplot as plt

sim = ah.Simulation(np.random.random((100, 6)), tau = 0.0, epsilon = 1e-1)

sim.start(1e3, save_to_array_every = 10)

fig, ax = plt.subplots()

ax.plot(sim.getEnergies())
ax.set_xlabel("Saved points")
ax.set_ylabel(r"Energy ($\frac{GmM}{r} + \frac{1}{2}mv^2$)")

fig.tight_layout()
plt.show()
